@@ -30,6 +30,7 @@ static dcp_dev_t *dcp;
 static dcp_iboot_if_t *iboot;
 static u64 fb_dva;
 static u64 fb_size;
+bool has_dcp;
 bool display_is_external;
 bool display_is_dptx;
 bool display_needs_power_cycle;
@@ -244,14 +245,23 @@ static uintptr_t display_map_fb(uintptr_t iova, u64 paddr, u64 size)
 
 const display_config_t *display_get_config(void)
 {
+    const display_config_t *conf = NULL;
+
     if (adt_is_compatible(adt, 0, "J473AP"))
-        return &display_config_m2;
+        conf = &display_config_m2;
     else if (adt_is_compatible(adt, 0, "J474sAP") || adt_is_compatible(adt, 0, "J475cAP"))
-        return &display_config_m2_pro_max;
+        conf = &display_config_m2_pro_max;
     else if (adt_is_compatible(adt, 0, "J180dAP") || adt_is_compatible(adt, 0, "J475dAP"))
-        return &display_config_m2_ultra;
+        conf = &display_config_m2_ultra;
     else
-        return &display_config_m1;
+        conf = &display_config_m1;
+
+    has_dcp = adt_path_offset(adt, conf->dcp) > 0;
+    if (!has_dcp) {
+        return NULL;
+    }
+
+    return conf;
 }
 
 int display_start_dcp(void)
@@ -266,6 +276,11 @@ int display_start_dcp(void)
 
     const display_config_t *disp_cfg = display_get_config();
 
+    if (!has_dcp) {
+        printf("display: device has no DCP. Display will not be initialised.\n");
+        return -1;
+    }
+
     display_is_dptx = !!disp_cfg->dptx_phy[0];
 
     dcp = dcp_init(disp_cfg);
@@ -601,6 +616,8 @@ int display_init(void)
         disp_path = "/arm-io/disp0";
 
     bool has_notch = false;
+    UNUSED(has_notch);
+
     int product = adt_path_offset(adt, "/product");
     if (product < 0) {
         printf("/product node not found!\n");
@@ -622,13 +639,14 @@ int display_init(void)
     else
         printf("display: Display is internal\n");
 
-    if (cur_boot_args.video.width == 640 && cur_boot_args.video.height == 1136) {
+    if ((cur_boot_args.video.width == 640 && cur_boot_args.video.height == 1136) &&
+        chip_id != S5L8960X) {
         printf("display: Dummy framebuffer found, initializing display\n");
         return display_configure(NULL);
-    } else if (display_is_external) {
+    } else if (display_is_external && is_mac) {
         printf("display: External display found, reconfiguring\n");
         return display_configure(NULL);
-    } else if (!(cur_boot_args.video.depth & FB_DEPTH_FLAG_RETINA)) {
+    } else if ((!(cur_boot_args.video.depth & FB_DEPTH_FLAG_RETINA)) && is_mac) {
         printf("display: Internal display with non-retina flag, assuming Sonoma bug and "
                "reconfiguring\n");
         fb_clear_direct(); // Old m1n1 stage1 ends up with an ugly logo situation, clear it.
@@ -649,6 +667,10 @@ int display_init(void)
 
 void display_shutdown(dcp_shutdown_mode mode)
 {
+    /* We have no DCP, so just exit */
+    if (!has_dcp)
+        return;
+
     if (iboot) {
         dcp_ib_shutdown(iboot);
         switch (mode) {
