@@ -3,6 +3,7 @@
 #ifndef DISPLAY_H
 #define DISPLAY_H
 
+#include "dcp.h"
 #include "types.h"
 
 typedef enum _dcp_shutdown_mode {
@@ -17,5 +18,6 @@ int display_init(void);
 int display_start_dcp(void);
 int display_configure(const char *config);
 void display_shutdown(dcp_shutdown_mode mode);
+const display_config_t *display_get_config(void);
 
 #endif
